"""
WalkingLiberty

Sort of a Bitcoin, Bitcoin Cash, and Bitcoin SV wallet API
"""

from walkingliberty.utilities import decode_uri

from hashlib import sha256

import bit
import bitcash
import bitsv

__version__ = '0.4.1'

VALID_CURRENCIES = ('btc', 'bch', 'bsv')
VALID_WALLET_MODES = ('deterministic-type1', 'wif', 'address')

DEFAULT_CURRENCY = VALID_CURRENCIES[0]
DEFAULT_WALLET_MODE = VALID_WALLET_MODES[0]


class WalkingLiberty():
    """
    WalkingLiberty class.
    """
    def __init__(self,
                 currency=None,
                 wallet_mode=None):
        # # We do this strange thing because if a caller sets None, we won't
        #  override it otherwise which will cause a needless failure.
        if currency is None:
            currency = DEFAULT_CURRENCY

        if wallet_mode is None:
            wallet_mode = DEFAULT_WALLET_MODE
        # #

        if currency not in VALID_CURRENCIES:
            message = 'currency must be one of: {}'.format(VALID_CURRENCIES)
            raise ValueError(message)

        if wallet_mode not in VALID_WALLET_MODES:
            message = 'wallet_mode must be one of: {}' \
                      ''.format(VALID_WALLET_MODES)
            raise ValueError(message)

        if currency == 'btc':
            self.bit = bit
        elif currency == 'bch':
            self.bit = bitcash
        elif currency == 'bsv':
            self.bit = bitsv

        self.currency = currency
        self.wallet_mode = wallet_mode

    def _validate_address(self, private_key):
        if self.currency == 'bsv':
            validator = self.bit.format.get_prefix
        else:
            validator = self.bit.format.get_version
        validator(private_key)

    def _private_key(self, private_key):
        """
        Returns a private key object
        """
        # sha256() wants bytes, Key() wants a string.
        if self.wallet_mode == 'deterministic-type1':
            private_key = bytes(private_key, 'utf-8')
            return self.bit.Key.from_hex(sha256(private_key).hexdigest())
        elif self.wallet_mode == 'wif':
            return self.bit.Key(private_key)
        elif self.wallet_mode == 'address':
            raise ValueError('Cannot derive private key from address.')

    def address(self, private_key):
        """
        Returns address for key
        """
        if self.wallet_mode == 'address':
            self._validate_address(private_key)
            return private_key
        private_key = self._private_key(private_key)
        return private_key.address

    def balance(self, private_key, unit='satoshi'):
        """
        Returns balance for a private key's address.

        Optionally, set unit for desired unit. (satoshi, btc, bsv, usd, etc)
        """
        if self.wallet_mode == 'address':
            self._validate_address(private_key)
            if self.currency == 'bsv':
                NetworkAPI = self.bit.network.NetworkAPI('main')
            else:
                NetworkAPI = self.bit.network.NetworkAPI

            satoshis = NetworkAPI.get_balance(address=private_key)
            satoshi_to_currency = self.bit.network.satoshi_to_currency_cached
            balance = satoshi_to_currency(satoshis, unit)
        else:
            private_key = self._private_key(private_key)
            # This is a formatted string, even if satoshis.
            balance = private_key.get_balance(currency=unit)

        # Return int if satoshis, str if anything else.
        if unit == 'satoshi':
            return int(balance)
        else:
            return balance

    def send(self, private_key, address, satoshis, unit='satoshi'):
        """
        Sends amount to address from the private_key's wallet.
        """
        if self.wallet_mode == 'address':
            raise ValueError('Cannot send without private key.')

        private_key = self._private_key(private_key)
        outputs = [(address, satoshis, unit)]
        # Have to call get_unspents() or get_balance() before send()
        private_key.get_unspents()
        return private_key.send(outputs, combine=False)

    def pay(self, private_key, uri):
        """
        Pays URI.
        """
        if self.wallet_mode == 'address':
            raise ValueError('Cannot pay without private key.')

        address, satoshis = decode_uri(self.currency, uri)
        private_key = self._private_key(private_key)
        outputs = [(address, satoshis, 'satoshi')]
        # Have to call get_unspents() or get_balance() before send()
        private_key.get_unspents()
        return private_key.send(outputs, combine=False)

    def sweep(self, private_key, address):
        """
        Sweeps all coins from private_key into address.
        """
        if self.wallet_mode == 'address':
            raise ValueError('Cannot sweep without private key.')

        private_key = self._private_key(private_key)
        private_key.get_unspents()
        return private_key.send([], combine=True, leftover=address)
